#include <Windows.h>
#include <Mstcpip.h>
#include <winsock2.h>
#include <ws2tcpip.h>
#include <stdio.h>

#include <RCF/RCF.hpp>
#include <RCF/TcpClientTransport.hpp>

#include <boost/function.hpp>
#include <boost/bind.hpp>

// Need to link with Ws2_32.lib
#pragma comment(lib, "ws2_32.lib")

RCF_BEGIN(Interface_Rcf, "Interface_Rcf")
   RCF_METHOD_V1(void, echo, const std::string&)
RCF_END(Interface_Rcf)


enum 
   {
   WM_IGI_SOCKETCLOSEDNOTIFIER = WM_USER
   };

const wchar_t g_szClassName[] = L"myWindowClass";
HWND hwnd;
int g_ConnectSocket;

class NotifierRegister
   {
   public:
   void registerForSocketNotification(int sock)
      {
      g_ConnectSocket = sock;
      //----------------------
       // Initialize Winsock
       WSADATA wsaData;
       int iResult = WSAStartup(MAKEWORD(2, 2), &wsaData);
       if (iResult != NO_ERROR) {
           wprintf(L"WSAStartup function failed with error: %d\n", iResult);
           return;
       }
       
      long evtInterestedIn;
      evtInterestedIn = FD_CLOSE;
      int ret = WSAAsyncSelect(sock, hwnd, WM_IGI_SOCKETCLOSEDNOTIFIER, evtInterestedIn);
      if (ret == SOCKET_ERROR)
         wprintf(L"socket notification was not able: %ld\n", WSAGetLastError());
      }
   };

LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
   {
      switch(msg)
      {
      case WM_IGI_SOCKETCLOSEDNOTIFIER:
         {
         wprintf(L"Received the WM_IGI_SOCKETNOTIFIER.");
         WSABUF buf;
         char wsaBuffer [4096] = {0};
         buf.buf = wsaBuffer;
         buf.len = 4096;
         DWORD flags = 0;
         DWORD bytesRead = 0;
         if (::WSARecv(g_ConnectSocket, &buf, 1, &bytesRead, &flags, 0,0) ==  SOCKET_ERROR)
            {
            int err = WSAGetLastError();
            switch (err)
               {
               case WSAEWOULDBLOCK:
                  wprintf(L"The connection would block.");
                  break;
               case WSAEBADF:
               case WSAEINVAL:
                  //error string is now set in read(), not here in nativeRead()
                  break;
               case WSAECONNRESET:
                  wprintf(L"The connection was resetted.");
                  break;
               case WSAECONNABORTED:
                  // for tcp sockets this will be handled in QNativeSocketEngine::read
                  wprintf(L"The connection was aborted.");
                  break;
               default:
                  break;
               }
            }
         break;
         }
      case WM_CLOSE:
         DestroyWindow(hwnd);
         break;
      case WM_DESTROY:
         PostQuitMessage(0);
         break;
      default:
         return DefWindowProc(hwnd, msg, wParam, lParam);
      }
      return 0;
   }

int main()
{
    WNDCLASSEX wc;
    MSG Msg;

    //Step 1: Registering the Window Class
    wc.cbSize        = sizeof(WNDCLASSEX);
    wc.style         = 0;
    wc.lpfnWndProc   = WndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = GetModuleHandle(0);
    wc.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
    wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
    wc.lpszMenuName  = NULL;
    wc.lpszClassName = g_szClassName;
    wc.hIconSm       = LoadIcon(NULL, IDI_APPLICATION);

    if(!RegisterClassEx(&wc))
    {
        MessageBox(NULL, L"Window Registration Failed!", L"Error!",
            MB_ICONEXCLAMATION | MB_OK);
        return 0;
    }

    // Step 2: Creating the Window
    hwnd = CreateWindowEx(
        WS_EX_CLIENTEDGE,
        g_szClassName,
        L"Client wsa notification",
        WS_OVERLAPPEDWINDOW,
        CW_USEDEFAULT, CW_USEDEFAULT, 240, 120,
        NULL, NULL, GetModuleHandle(0), NULL);

    if(hwnd == NULL)
    {
        MessageBox(NULL, L"Window Creation Failed!", L"Error!",
            MB_ICONEXCLAMATION | MB_OK);
        return 0;
    }

    ShowWindow(hwnd, SW_SHOWDEFAULT);
    UpdateWindow(hwnd);

   RCF::RcfInitDeinit rcfInit;
   RcfClient<Interface_Rcf> client(RCF::TcpEndpoint("192.168.2.232", 50001));

   //set the after-socket-creation-before-connect callback.
   NotifierRegister notifierRegister;
   RCF::TransportType transType = client.getClientStub().getTransport().getTransportType();
   RCF::TcpClientTransport& trans = dynamic_cast<RCF::TcpClientTransport&>(client.getClientStub().getTransport());
   boost::function<void(int)> callback(boost::bind(&NotifierRegister::registerForSocketNotification,
      &notifierRegister, ::_1));
   RCF::IpAddress addr = trans.getRemoteAddr();
   addr.setAfterCreateSocketCallback(callback);
   trans.setRemoteAddr(addr);
               
   client.getClientStub().connect();

   //enable TCP keep-alive
   RCF::TcpClientTransport &tcpClientTransport = dynamic_cast<RCF::TcpClientTransport &> (
      client.getClientStub().getTransport());
   int sock = tcpClientTransport.getNativeHandle();
   struct tcp_keepalive keepAlive = {0};
   keepAlive.onoff = TRUE;
   keepAlive.keepalivetime = 2000;
   keepAlive.keepaliveinterval = (3000 / 10);
   DWORD bytesReturned = 0;
   WSAIoctl(sock, SIO_KEEPALIVE_VALS, &keepAlive, sizeof(keepAlive), NULL, 0, &bytesReturned, NULL, NULL);

   //send a test message over the altered socked.
   client.echo("Echo from client");

    // Step 3: The Message Loop
    while(GetMessage(&Msg, NULL, 0, 0) > 0)
    {
        TranslateMessage(&Msg);
        DispatchMessage(&Msg);
    }
    return Msg.wParam;
}
